#!/usr/bin/wish

namespace eval ::calc {
  namespace export calc ;# public interface
}

proc ::calc::bindings {mode} {
  # bindings copied from ngcgui_ttt.tcl
  switch $mode {
    setup {
     ::calc::bind_for_axis $::clc(topf)
     bindtags $::clc(topf) $::clc(topf)
     entry_mend $::clc(topf)
     bind $::clc(topf) <Enter> [list ::calc::bindings enter]
     bind $::clc(topf) <Leave> [list ::calc::bindings leave]
     set ::clc(restore,focus) [focus -lastfor $::clc(topf)]
    }
    enter {
      set ::clc(restore,focus) [focus -lastfor $::clc(topf)]
      focus $::clc(topf)
    }
    leave {
     focus -force $::clc(restore,focus)
    }
  }
} ;# bindings

proc ::calc::bind_for_axis {w} {
  # bind_for_axis copied from ngcgui.tcl

  # Escape and other special bindings for axis embedding
  bind $w <Key-Escape> "$::clc(parent) raise preview"

  # axis omits return break in estopped_clicked for F1
  bind $w <Key-F1> "[bind all <Key-F1>];break"

  # Fn keys
  foreach i {2 3 4 5 6 7 8 9 10 11 12} {
    bind $w <Key-F$i> "[bind   . <Key-F$i>];break"
  }
} ;# bind_for_axis

proc ::calc::entry_mend {w} {
  # entry_mend copied from ngcgui.tcl
  # axis creates jog bindings for the toplevel (.==dot):
  # for <KeyPress-minus> <KeyRelease-minus>  <KeyPress-equal> <KeyRelease-equal>
  # eg: bind . <KeyPress-minus> ==> {stuff}
  # thus, for entries, bindtags are: {$e Entry . all} <-- the . is a problem
  # so, limit the bindtags for entries
  if {[winfo class $w] == "Entry"} {
    bindtags $w [list $w Entry all] ;# remove the . bindtag
    bind_for_axis $w
  }
  foreach child [winfo children $w] {
    if {$child == ""} continue
    ::calc::entry_mend $child
  }
} ;# entry_mend

proc ::calc::embedinit {} {
  if [info exists ::clc(inhibit)] return ;# provision for non-axis sourcing
  # init for embedding
  set ::clc(info) "Mill Calculator testing embed in axis"
  set ::clc(pagename) calc

  set wtab [dynamic_tab $::clc(pagename) $::clc(pagename)] ;# axis function
  set ::clc(topf) [frame $wtab.[qid] -container 0 -borderwidth 0 -highlightthickness 0]
  pack $::clc(topf) -side top -fill both -expand 1 -anchor nw

  set ::clc(parent) [winfo parent $wtab]

  # use ::ngcgui::tabmanage if available:
  if {[info procs ::ngcgui::tabmanage] == "::ngcgui::tabmanage"} {
    ::ngcgui::tabmanage $::clc(pagename) $::clc(topf) \
                        "Calculator"\
                        ::clc(info)
    set w [frame $::clc(topf).t]
    pack $w -anchor nw -expand 0 -fill x -side top
    set cframe $w
  } else {
    set cframe $::clc(topf)
  }

  ::calc::preset ::clc
  # specific settings for embedding herein
  set ::clc(aspect)   horiz

  set  w [::calc::gui create $cframe.calc]
  if {"$w" == ""} {
    puts stderr "::calc::embedinit: problem with ::calc::gui create"
  } else  {
    pack $w -anchor nw -side top -fill x -expand 1
  }
  ::calc::bindings setup
  $::clc(parent) raise preview
} ;# embedinit

proc ::calc::preset {ay_name} {
  # using apps call this to populate ay_name,
  # superseded items as reqd
  # all required items with defaults:
  upvar $ay_name ay
  set ay(pi) 3.14159
  set ay(color,computed) green4
  set ay(color,user)     blue
  set ay(color,pending)  purple
  set ay(color,default)  black
# consider also culength/ipm = timetocut
  set ay(items) [list DIAM "%.3f" "Diameter"\
                      NF   "%2d"  "Number of flutes"\
                      IPT  "%.4f" "Feed per tooth"\
                      IPR  "%.4f" "Inch per revolution"\
                      SFM  "%.0f" "Surface feet per minute"\
                      RPM  "%.0f" "Spindle speed"\
                      IPM  "%.1f" "Table feed"\
                      RDC  "%.3f" "Radial depth of cut (width)"\
                      ADC  "%.3f" "Axial depth of cut (length)"\
                      MRR  "%.3f" "Metal removal rate (in^3/min"\
                 ]
  # startup values:
  set ay(DIAM)   1.0
  set ay(NF)     2
  set ay(IPT)    0.001
  set ay(SFM)  300
} ;# preset

proc ::calc::gui {mode args} {
  # use ::calc::preset for required ::clc() items and defaults
  # standalone invoke: ::calc::gui standalone wframe
  # embedded   invoke: ::calc::gui create     wframe
  switch $mode {
    standalone {
      set ::clc(standalone) 1
      set w [::calc::gui create $args]
      return $w
    }
    create {
      # mandatory arg for mode==create
      set wframe [lindex $args 0]
      if {"$wframe" == ""} {
        return -code error "gui:create no arg for wframe"
      }
      # caller packs/unpacks wframe which must be a valid name
      # but not exist yet
      set w [frame $wframe -bd 0] ;# wframe specifies name, create it here
      set ::clc(wframe) $wframe
      set ::clc(top) [winfo toplevel $wframe]

      # defaults:
      set ::clc(id)      0

      # initializations:
      foreach {item fmt txt} $::clc(items) {
        set ::clc($item,last) ""
        set ::clc($item,fmt)  "$fmt"
      }

      set bw 28
      switch $::clc(aspect) {
        vert {
          set wI [frame $wframe.input   -bd 2 -relief sunken] ;# input frame
          set wE [frame $wframe.exit    -bd 2 -relief sunken] ;# exit frame

          pack $wI -side top    -fill x -expand 1 -anchor n
          pack $wE -side bottom -fill x -expand 1 -anchor s
          set ::clc(pack,positional) top
          set ::clc(pack,global)     top
        }
        horiz {
          set wL [frame $wframe.left -bd 2 -relief sunken] ;# left frame
          set wI [frame $wL.input    -bd 2 -relief sunken] ;# input frame
          set wE [frame $wL.exit     -bd 2 -relief sunken] ;# exit frame

          pack $wL -side left -fill x -expand 1 -anchor n
          pack $wI -side top  -fill x -expand 1 -anchor n
          pack $wE -side bottom  -fill x -expand 1 -anchor s
          set ::clc(pack,positional) left
          set ::clc(pack,global)     left
        }
        default {return -code error calc::gui:aspect <$aspect>}
      }
#     set ::clc(varframe) $wV
      set ::clc(iframe)   $wI

      set w [frame $wI.[qid]]
      pack $w -fill x -expand 1
      pack [label $w.[qid] -anchor w -text "Feed and Speed" \
          -bg lightsteelblue2 -relief groove] -fill x -expand 1

      foreach {item fmt txt} $::clc(items) {
        set w [frame $wI.[qid]]
        pack $w -fill x -expand 1
        set l [label $w.[qid] -relief groove -anchor e\
              -width $bw -text $txt ]
        pack $l -side left -expand 0 -fill none

        set l [label $w.[qid] -relief groove -anchor e\
              -width 5 -text $item ]
        pack $l -side left -expand 0 -fill x

        set tvar ::clc($item)
        set e [entry $w.[qid] -textvariable $tvar \
              -width 8\
              -validate     all\
              -validatecommand [list ::calc::validateNumber $item $tvar %W %s %P]]
        set ::clc($item,entrywidget) $e
        pack $e -side left -expand 1 -fill x
        bind $e <Return> [list ::calc::userchanged $item]
      }

      set ::clc(msg,widget) [label $wE.[qid] -anchor w] ;# update with config
      pack $::clc(msg,widget) -side left -fill x -expand 1
      $::clc(msg,widget) configure -text \
         "Enter values, then <RETURN> to (re) compute"

      if [info exists ::clc(standalone)] {
        set b [button $wE.[qid] -takefocus 1\
               -text "Exit" -width 4 -command ::calc::bye]
        pack $b -side left -fill x -expand 0
      }

      return $wframe
    }
    default {return -code error "::calc::gui: unknown mode <$mode>"}
  }
  puts stderr "NOTREACHED reached mode=<$mode>"
} ;# gui

proc ::calc::userchanged {item} {
  colors user $item
  switch $item {
    NF      -
    IPT     {compute IPR; compute IPM}
    IPR     {compute IPT; compute IPM}
    DIAM    -
    SFM     {compute RPM; compute IPM; compute IPR}
    RPM     {compute SFM; compute IPM}
    IPM     {compute ipm-IPR; compute IPT}
    MRR     {compute IPM}
    RDC     {}
    ADC     {}
    default {puts "calc::userchanged:unknown_item<$item>"}
  }
  compute MRR ;# always
} ;# userchanged

proc ::calc::colors {mode i} {
  #puts "colors $mode $i"
  foreach {item fmt txt} $::clc(items) {
    switch $mode {
      user {
        if {"$i" == "$item"} {
          set ::clc($item,last) user
          $::clc($item,entrywidget) conf -fg $::clc(color,user)
        } else {
          $::clc($item,entrywidget) conf -fg $::clc(color,default)
        }
      }
      computed {
        if {"$i" == "$item"} {
          set ::clc($item,last) computed
          $::clc($item,entrywidget) conf -fg $::clc(color,computed)
        }
      }
    }
  }
} ;# colors

# Cutting Speed:              SFM      = DIAM x pi/12 x RPM
# Spindle Speed:              RPM      = SFM x 12/pi ÷ DIAM
# Table Feed:                 IPM      = IPT x NF x RPM

# Feed per Tooth:             IPT      = IPM ÷ NF ÷ RPM
# Inch (Feed) per Revolution: IPR (FR) = IPM ÷ RPM
# Metal Removal cubic in/min: MRR      = IPM x RDC x ADC
#
# D:   inch      Diameter of Tool.
# NF:            Number of Flutes.
# FR:  in/rev    Feed per Revolution.
# IPM: in/min    Table Feed.
# IPR: in/rev    Inch per Revolution.
# IPT: in/tooth  Feed per Tooth.
# SFM: ft/min    Cutting Speed.
# RPM: rev/min   Spindle Speed.
# ADC:           Axial depth-of-cut (Length).
# HP:            Actual Horsepower available at running RPM.
#                HP=MRR*mf
#                   mf = 1.0  steel
#                   mf = 0.65 gray iron
#                   mf = 0.3  aluminum
# PC:            Power constants for HP (HP/CI-MR).
# RDC:           Radial depth-of-cut (Width).
# MRR:           Metal Removal Rate (Cubic inches per minute)
proc ::calc::compute {item} {
  #puts ::calc::compute<$item>
  set err 0
  switch $item {
    IPR {
      if [catch {set ::clc(IPR) \
         [expr 1.0*$::clc(NF)*$::clc(IPT)]} msg] {set err 1}
    }
    ipm-IPR {
      if [catch {set ::clc(IPR) \
         [expr 1.0*$::clc(IPM)/$::clc(RPM)]} msg] {set err 1}
         set item IPR
    }
    IPT {
      if [catch {set ::clc(IPT) \
         [expr 1.0*$::clc(IPR)/$::clc(NF)]} msg] {set err 1}
    }
    SFM {
      if [catch {set ::clc(SFM) \
         [expr 1.0*$::clc(pi)/12.*$::clc(DIAM)*$::clc(RPM)]} msg] {set err 1}
    }
    RPM {
      if [catch {set ::clc(RPM) \
         [expr 12.0/$::clc(pi)*$::clc(SFM)/$::clc(DIAM)]} msg] {set err 1}
    }
    IPM {
      if [catch {set ::clc(IPM) \
         [expr 1.0*$::clc(IPT)*$::clc(NF)*$::clc(RPM)]} msg] {set err 1}
    }
    MRR {
      if [catch {set ::clc(MRR) \
         [expr 1.0*$::clc(IPM)*$::clc(RDC)*$::clc(ADC)]} msg] {set err 1}
    }
    RDC -
    ADC {}
    default {puts unknown_item}
  }

  # if $err {puts "::calc::compute:err:<$item><$msg>"}
  if $err {return 0}

  if {"$::clc($item)" != ""} {
    set ::clc($item) [format $::clc($item,fmt) $::clc($item)]
    #puts "new $item: $::clc($item)"
  }

  $::clc(msg,widget) conf -text "Computed $item = $::clc($item)"
  colors computed $item
  return 1
} ;# compute

proc ::calc::qid {} {
  # unique identifier
  if ![info exists ::clc(qid)] { set ::clc(qid) 0 }
  return [incr ::clc(qid)]
} ;# qid

proc ::calc::bye {} {
  catch {after cancel $::clc(afterid)}
  destroy $::clc(top)   ;# for embedded usage
  set ::calc::finis 1 ;# for standalone usage
} ;# bye

proc ::calc::top {ay_name} {
  # make a standalone toplevel
  upvar $ay_name ay
  toplevel    .calc
  wm protocol .calc WM_DELETE_WINDOW ::calc::bye

  foreach n [array names ay] { set ::clc($n) $ay($n) }

  set w [::calc::gui standalone .calc.w]
  compute IPR; compute RPM; compute IPM
  if {"$w" == ""} {exit 1} ;# "" indicates something went wrong
  pack $w
} ;# top


proc ::calc::validateNumber {item varname widget current new} {
  # all entries must be numbers
  if ![info exists $varname] {return 1}
  if [catch  {format %g $new} ] {
    $widget configure -fg red
    return 1 ;# problem but return ok (just change color)
  } else {
    if {"$current" != "$new"} {
      $widget configure -fg $::clc(color,pending)
    } else {
      switch $::clc($item,last) {
        user     {$widget configure -fg $::clc(color,user)}
        computed {$widget configure -fg $::clc(color,computed)}
        default  {$widget configure -fg $::clc(color,default)}
      }
    }
    return 1 ;# 1==>ok
  }
} ;# validateNumber

proc ::calc::tst {} {
  # to debug using tkcon: source this file then % ::calc::tst
  package require Tk
  catch {destroy .calc}
  ::calc::preset ::clc  ;# setup ::clc() with defaults
  set control(aspect)     vert
  set control(font)       {Helvetica -16 bold}
  set control(app)        [file tail $::argv0]
  set control(app)        tst ;# tkcon argv0 not available

  option add *Font $control(font)

  eval ::calc::top control
  tkwait variable ::calc::finis
} ;# tst

proc ::calc::usage {ay_name} {
  upvar $ay_name ay
  set prog [file tail $::argv0]
  set dfont "\"$ay(font)\"" ;# avoid messing up vim colors
  puts stderr "Usage:
  $prog  --help
  $prog \[Options\]

  Options:
         \[tbd]\
"
  exit 0
} ;# usage

if [catch {
  # configure standalone usage:
  if 0 {
    puts "::clc:debugging begin"
    foreach v [info vars] {
      if [array exists $v] {
        parray $v
      } else {
        puts "$v=[set $v]"
      }
    }
    puts "::clc:debugging end"
  } ;# if 0/1

  if {[info exists ::argv0] && [info script] == $::argv0} {
    set control(axis,embed) 0
    ::calc::preset control ;# setup control() with defaults
    wm withdraw . ;# use a named toplevel
    # configure for standalone usage:
    set control(font)       {Helvetica -16 bold} ;# default
    set control(aspect)     vert
    set control(app)        [file tail $argv0]

    while {[llength $::argv] >0} {
      # beware wish handling of reserved cmdline arguments
      # to use -h: use -- -h,
      # lreplace shifts argv for no. of items for each iteration
      switch -- [lindex $::argv 0] {
        -h     -
        --help {::calc::usage control;exit 0}
        -horiz {set control(aspect) horiz
                set ::argv [lreplace $::argv 0 0]
               }
        -vert  {set control(aspect) vert
                set ::argv [lreplace $::argv 0 0]
               }
        -font  {set control(font) [lindex $::argv 1]
                set ::argv [lreplace $::argv 0 1]
               }
        -i          -
        --input     -
        --subfile   {set control(filename,subfile) [lindex $::argv 1]
                     set ::argv [lreplace $::argv 0 1]
                    }

        default {break}
      }
    }
    switch -- $control(font) {
      small   {option add *Font {Helvetica -10 bold}}
      big     {option add *Font {Helvetica -16 bold}}
      default {option add *Font $control(font)}
    }
    eval ::calc::top control ;# control() specifies args
    wm withdraw . ;# sometimes early withdraw doesnt work
    tkwait variable ::calc::finis
    exit 0
  }
} msg] {puts "calc.tcl:msg=$msg"}

# make sure this file doesn't restart embedding if sourced multiple times
if {   ![info exists ::clc(axis,embed)] \
    && ([info commands "dynamic_tab"] == "dynamic_tab") } {
  set ::clc(axis,embed) 1
  ::calc::embedinit ;# embed in axis
}
package provide Calc 1.0
